import numpy as np

import matplotlib.pyplot as plt


from scipy.optimize import curve_fit
from scipy.constants import physical_constants


# Erklärung:
# Importiere uncertainties unter dem Namen unc
# Importiere uncertainties.unumpy unter dem Namen unp
# Importiere der Funktionen nominal_values und std_devs
# aus uncertainties.unumpy, unter den kürzeren Namen
# noms respektive stds
import uncertainties as unc
import uncertainties.unumpy as unp
from uncertainties.unumpy import (nominal_values as noms, std_devs as stds)

from curve_fit import ucurve_fit
from latex_formatting import make_qty

#
# Erklärung:
# Die Dateien im Ordner data enthalten jetzt auch Spalten mit den Unsicherheiten
# für die meisten Messwerte.
# Beim Importieren der Daten muss beachtet werden, dass:
# (1) Messgröße und Unsicherheit jeweils eine eigene Variablennamen brauchen.
# Auch hier ist eine konsistente Benennung sinnvoll
# z.B. l (Messwerte) und l_unc (zugehörige Unsicherheiten)
#
# (2) die Unsicherheitbehafteten Messwerte noch erstellt werden müssen,
# entweder durch unc.ufloat oder durch das unp.uarray


# Länge der schiefen Ebene
l, l_unc = np.genfromtxt("data/Messwerte_Bahn.txt", delimiter=",", unpack=True)/100 # m
l = unc.ufloat(l, l_unc)

# Framerate der Kamera (hat keine Unsicherheit, fps_unc = 0)
fps, fps_unc = np.genfromtxt("data/Messwerte_Kamera.txt", delimiter=",", unpack=True) # 1/s

# Masse und Umfang der Kugel
m_b, m_b_unc, u_b, u_b_unc = np.genfromtxt("data/Messwerte_Kugel.txt", delimiter=",", unpack=True)
m_b = unc.ufloat(m_b,m_b_unc)/1000 # kg
u_b = unc.ufloat(u_b,m_b_unc)/100 # m

# Messreihe: Starthöhe und Startframe und Endframe (Kugel)
h_b, h_b_unc, Fi_b, Fi_b_unc, Ff_b, Ff_b_unc = np.genfromtxt("data/Messwerte_Frames_Kugel.txt", delimiter=",", unpack=True)
h_b = unp.uarray(h_b, h_b_unc)/100 # m
Fi_b = unp.uarray(Fi_b, Fi_b_unc)
Ff_b = unp.uarray(Ff_b, Ff_b_unc)


# Masse und Umfang des Zylinders
m_c, m_c_unc, u_c, u_c_unc, d_c, d_c_unc = np.genfromtxt("data/Messwerte_Zylinder.txt", delimiter=",", unpack=True)
m_c = unc.ufloat(m_c, m_c_unc)/1000 # kg
u_c = unc.ufloat(u_c, u_c_unc)/100 # m
d_c = unc.ufloat(d_c, d_c_unc)/100 # m

# Messreihe: Starthöhe und Startframe und Endframe (Zylinder)
h_c, h_c_unc, Fi_c, Fi_c_unc, Ff_c, Ff_c_unc = np.genfromtxt("data/Messwerte_Frames_Zylinder.txt", delimiter=",", unpack=True)
h_c = unp.uarray(h_c, h_c_unc)/100 # m
Fi_c = unp.uarray(Fi_c, Fi_c_unc)
Ff_c = unp.uarray(Ff_c, Ff_c_unc)


# Die folgenden Berechnungen funktionieren alle weiterhin, nur jetzt mit automatischer Fehlerrechnung


# Berechnung der benötigten Größen: Radius und Trägheitsmoment

# Radius der Kugel
r_b = u_b/(2*np.pi)

# Äußerer Radius des Zylinders
ro_c = u_c/(2*np.pi)

# Innerer Radius des Zylinders
ri_c = ro_c - d_c


# Trägheitsmoment der Kugel
I_b = 2/5 * m_b * r_b**2

# Trägheitsmoment des Zylinders
I_c = 1/2 * m_c * (ro_c**2 + ri_c**2)

print("Trägheitsmoment (Kugel)")
print(I_b)

print("Trägheitsmoment (Zylinder)")
print(I_c)


# Berechnung der Rollzeiten aus den Messwerten der Frames und Framerate
t_b = (Ff_b - Fi_b)/fps
t_c = (Ff_c - Fi_c)/fps

# Berechnung des Mittelwerts für die Zeitdauer t für jede (dreifach gemessene) Höhe


t_b_mean = t_b.reshape(-1,3).mean(axis=1)


h_b_mean = h_b.reshape(-1,3)[:,0]


t_c_mean = t_c.reshape(-1,3).mean(axis=1)
h_c_mean = h_c.reshape(-1,3)[:,0]

# Ausgabe verarbeiteten der Messwerte

print("Messwerte (Kugel)")
print("alle Zeiten")
print(t_b)
print("alle Höhen")
print(h_b)
print("Höhe")
print(h_b_mean)
print("gemittlelte Zeit")
print(t_b_mean)
print("\n")


print("Messwerte (Zylinder)")
print("alle Höhen")
print(h_c)
print("alle Zeiten")
print(t_c)
print("Höhe")
print(h_c_mean)
print("gemittlelte Zeit")
print(t_c_mean)
print("\n")

# Erklärung:
# Alle Ergebnisse die dieses Python-Script berechnet sollten
# für eine noch umfangreichere Automatisierung nicht mehr
# per Hand in das Protokoll geschrieben werden.
# Neben der Automatisierung verhindert dies auch Fehler
# beim Abtippen oder durch vergessene Aktualisierungen,
# falls sich die Werte im Verlauf der Bearbeitungszeit
# nocheinmal ändern.
#
# Die Ausgabe der Werte aufs Terminal erspart das nachschauen
# in den Dateien. Es ist aber zu empfehlen, die Speicherung
# in die Dateien  unter der Ausgabe zu machen, um Veränderungen
# der Werte zwischen Ausgabe aufs Terminal und Speichern vorzubeugen.
# nocheinmal ändern.
# Ausgenommen davon die Tabellen, weil der Code recht viel Platz einnimmt,
# werden diese am Ende der Datei exportiert.
#

g = physical_constants["standard acceleration of gravity"][0]
with open("build/theoretical-g.tex", "w") as f:
    f.write(make_qty(g,r"\meter\per\second\squared", figures=2, formatting="per-mode=reciprocal"))

with open("build/framerate.tex", "w") as f:
    f.write(make_qty(fps, r"\per\second", figures=0, formatting="per-mode=reciprocal"))

with open("build/theoretical-I_ball.tex", "w") as f:
    f.write(make_qty(I_b*10000,r"\kilo\gram\centi\meter\squared")) # change units: kg cm²

with open("build/theoretical-I_cylinder.tex", "w") as f:
    f.write(make_qty(I_c*10000,r"\kilo\gram\centi\meter\squared"))# change units: kg cm²

with open("build/tracklength.tex", "w") as f:
    f.write(make_qty(l,r"\meter"))

with open("build/mass_ball.tex", "w") as f:
    f.write(make_qty(m_b,r"\kilo\gram"))

with open("build/mass_cylinder.tex", "w") as f:
    f.write(make_qty(m_b,r"\kilo\gram"))

with open("build/radius_ball.tex", "w") as f:
    f.write(make_qty(r_b, r"\meter"))

with open("build/radius-inner_cylinder.tex", "w") as f:
    f.write(make_qty(ri_c,r"\meter"))

with open("build/radius-outer_cylinder.tex", "w") as f:
    f.write(make_qty(ro_c,r"\meter"))

# Erstelle Plots der Messwerte t_._mean und h_._mean








# Erklärung:
# Nur für die Erstellung der Plots und Fits ändert sich etwas:
# Weder scipy noch matplotlib können direkt mit den unsicherheitbehafteten
# Messwerten umgehen.
# Folgendes muss dafür geändert werden:
#
# (1) in der Fitfunktion müssen mit der Funktion noms()
#
# die Unsicherheiten von Konstanten entfernt werden
#
# (2) mit Unsicherheiten der unabhängigen Variable kann curve_fit nicht umgehen,
# diese müssen mit noms() entfernt werden.
#
# (3) die Werte und Unsicherheiten der abhängigen Variable (gemessene Funktionswerte)
# müssen getrennt übergeben werden: noms() und stds()
#
# (4) die Darstellung der Messwerte im Plot wird durch errorbar() ersetzt,
# um die Unsicherheiten anzeigen zu können
#
# In der Aufgabe 3-curve_fit muss eine Funktion ucurve_fit geschrieben werden,
# die die scipy Funktion curve_fit verwendet, die aber den Umgang mit den
# Unsicherheiten abstrahiert, sodass man das nicht jedes Mal aufs neue machen muss.

def fit_g_ball(h, g, t0):
    return np.sqrt(7/5 * 1/h * 2* noms(l)**2/g) + t0

def fit_g_cylinder(h, g, t0):
    return np.sqrt((3 + noms(ri_c**2/ro_c**2)) * noms(l)**2/g * 1/h) + t0



# Berechnung der Fitparameter für die Kugel


# Erklärung:
# Wir verwenden hier den 'Wrapper' für curve_fit, aus Aufgabe 5-uncertainties/3-curve_fit,
# um die Arbeit mit unsicherheitbehafteten Werten für Fits wiederverwendbar auszulagern.
params = ucurve_fit(fit_g_ball, noms(h_b_mean), t_b_mean)

print("Fitparameter (Kugel)")
for name, param in zip("gt", params):
    print(f"{name} = {param.n:8.3f} ± {param.s:.3f}")

# exporting fit parameters
parameter_g, parameter_t0 = params

with open("build/parameter-g_kugel.tex", "w") as f:
     f.write(make_qty(parameter_g, r"\meter\per\second\squared"))

with open("build/parameter-t0-g_kugel.tex", "w") as f:
     f.write(make_qty(parameter_t0, r"\second"))

h_plot = np.linspace(0.03, 0.33, 205)


# Fitparameter für den Zylinder


fig, ax = plt.subplots(1, 1, layout="constrained")


ax.set_xlabel("$h$ / m")
ax.set_ylabel("$t$ / s")



ax.errorbar(noms(h_b_mean), noms(t_b_mean), yerr=stds(t_b_mean), fmt="k+", label="Daten: Kugel")

ax.plot(h_plot, fit_g_ball(h_plot, *noms(params)),  label="Fit")
ax.legend()
fig.savefig("build/plot-g_kugel.pdf")



# Berechnung der Fitparameter für den Zylinder

params = ucurve_fit(fit_g_cylinder, noms(h_c_mean), t_c_mean)

print("Fitparameter (Zylinder)")
for name, param in zip("gt", params):
    print(f"{name} = {param.n:8.3f} ± {param.s:.3f}")
print("\n")

 # exporting fit parameters
parameter_g, parameter_t0 = params

with open("build/parameter-g_zylinder.tex", "w") as f:
     f.write(make_qty(parameter_g, r"\meter\per\second\squared"))

with open("build/parameter-t0-g_zylinder.tex", "w") as f:
     f.write(make_qty(parameter_t0, r"\second"))

fig, ax = plt.subplots(1, 1, layout="constrained")

ax.set_xlabel("$h$ / m")
ax.set_ylabel("$t$ / s")


ax.errorbar(noms(h_c_mean), noms(t_c_mean), yerr=stds(t_c_mean), fmt="k+", label="Daten: Zylinder")
ax.plot(h_plot, fit_g_cylinder(h_plot, *noms(params)),  label="Fit")
ax.legend()
fig.savefig("build/plot-g_zylinder.pdf")




def fit_I_ball(h, I, t0):
    g = physical_constants["standard acceleration of gravity"][0]
    return np.sqrt(2/g * 1/h * noms(l)**2 * (1 + I/(noms(m_b * r_b)**2)) ) + t0


def fit_I_cylinder(h, I, t0):
    g = physical_constants["standard acceleration of gravity"][0]
    return np.sqrt(2/g * 1/h * noms(l)**2 * (1 + I/(noms(m_c * ro_c**2)) )) + t0



# Berechnung der Fitparameter für die Kugel

params = ucurve_fit(fit_I_ball, noms(h_b_mean), t_b_mean, bounds=([0,-np.inf],[+np.inf,+np.inf]))

print("Fitparameter (Kugel)")
for name, param in zip("It", params):
    # in kg*m² sehr klein, umwandlung zu kg*cm²
    if name == "I":
        param *= 10000
    print(f"{name} = {param.n:8.3f} ± {param.s:.3f}")
print("\n")
 # exporting fit parameters
parameter_I, parameter_t0 = params

with open("build/parameter-I_kugel.tex", "w") as f:
     # in kg*m² sehr klein, umwandlung zu kg*cm²
     f.write(make_qty(parameter_I*10000, r"\kilo\gram\centi\meter\squared"))

with open("build/parameter-t0-I_kugel.tex", "w") as f:
     f.write(make_qty(parameter_t0, r"\second"))


fig, ax = plt.subplots(1, 1, layout="constrained")

ax.set_xlabel("$h$ / m")
ax.set_ylabel("$t$ / s")


ax.errorbar(noms(h_b_mean), noms(t_b_mean), yerr=stds(t_b_mean), fmt="k+", label="Daten: Kugel")
ax.plot(h_plot, fit_I_ball(h_plot, *noms(params)),  label="Fit")
ax.legend()
fig.savefig("build/plot-I_kugel.pdf")


# Berechnung der Fitparameter für den Zylinder

params = ucurve_fit(fit_I_cylinder, noms(h_c_mean), t_c_mean, bounds=([0,-np.inf],[+np.inf,+np.inf]))

for name, param in zip("It", params):
    # in kg*m² sehr klein, umwandlung zu kg*cm²
    if name == "I":
        param *= 10000
    print(f"{name} = {param.n:8.3f} ± {param.s:.3f}")
print("\n")

# exporting fit parameters
parameter_I, parameter_t0 = params

with open("build/parameter-I_zylinder.tex", "w") as f:
     # in kg*m² sehr klein, umwandlung zu kg*cm²
     f.write(make_qty(parameter_I*10000, r"\kilo\gram\centi\meter\squared"))

with open("build/parameter-t0-I_zylinder.tex", "w") as f:
     f.write(make_qty(parameter_t0, r"\second"))


fig, ax = plt.subplots(1, 1, layout="constrained")

ax.set_xlabel("$h$ / m")
ax.set_ylabel("$t$ / s")


ax.errorbar(noms(h_c_mean), noms(t_c_mean), yerr=stds(t_c_mean), fmt="k+", label="Daten: Zylinder")
ax.plot(h_plot, fit_I_cylinder(h_plot, *noms(params)),  label="Fit")
ax.legend()
fig.savefig("build/plot-I_zylinder.pdf")
# writing latex tables
# first: writing all measured values not just the averages
# heights_ball, durations_ball, durations_cylinder
# have not been reduced to 6 values each

# Erklärung:
# Die Erstellung der Tabellen wird hier nicht zu 100% automatisiert, aber
# zumindest das abtippen der Werte ist hiermit nicht mehr nötig.
# Das ist auch bei Weitem nicht nur ein Ausdruck von Faulheit,
# sondern wieder ein wichtiger Schritt zu guter wissenschaftlicher
# Praxis: Messwerte per Hand abtippen ist fehleranfällig und nur
# mäßig reproduzierbar.
# Es steht euch frei das noch etwas mehr zu automatisieren.
#
# Jede Tabelle besteht aus drei Teilen, dem
# i) table_header: Beginn der Latexumgebung tblr und alle Spalten/Zeileneinstellungen
# sowie die Kopfzeilen der Tabelle
#
# iii) table_footer: \bottomrule und das Ende der Latexumgebung tblr
#
# ii) row_template: Ein string mit Format angaben für die Werte in jeder Spalte,
# (die Formatangaben stimmen nur hinter dem . mit denen für tblr überein
#  z.B. Spalte 2: table-format=4.0 aber {1.n: 0.0f})
# Die Zahl am Anfang der Formatangabe legt die Reihenfolge fest,
# das .n bzw. .s bezeichnet bei unsicherheitbehafteten Werten die
# nominalen Werte respektive die Standardabweichung.
# .n und .s Formatangaben mit der selben Zahl davor sollten immer aufeinander folgen.
#
# Die drei Teile werden dann in einer Datei zusammen gesetzt:
# Zuerst wird der table_header in die Datei geschrieben
# dann werden die Daten, mit Hilfe einer for-loop über alle Spalten
# (beachtet die Verwendung von zip) und dem row_template,
# Zeile für Zeile in die Datei geschrieben.
# Danach muss mit f.write('\n') dafür gesorgt werden, dass in der Datei eine
# neue Zeile angefangen wird.
# Und am Ende wird noch der table_footer in die Datei geschrieben.
#

table_header = r"""
  \begin{tblr}{
    colspec = {
        S[table-format=1.3]
        S[table-format=4.0]
        S[table-format=1.0]
        S[table-format=4.0]
        S[table-format=1.0]
        S[table-format=1.3]
        S[table-format=1.3]
        S[table-format=4.0]
        S[table-format=1.0]
        S[table-format=4.0]
        S[table-format=1.0]
        S[table-format=1.3]
        S[table-format=1.3]
    },
    row{1} = {guard},
    row{2} = {guard, mode=math},
    vline{3,5,7,9,11,13} = {2}{-}{text=\clap{\pm}},
  }
    \toprule
    &
    \SetCell[c=6]{c} Messung: Kugel & & & & & &
    \SetCell[c=6]{c} Messung: Zylinder & & & & & \\
    \cmidrule[lr]{2-7}\cmidrule[lr]{8-14}
    h \mathbin{/} \unit{\meter} &%
    \SetCell[c=2]{c} F_\text{i} \mathbin{/} 1 & &%
    \SetCell[c=2]{c} F_\text{f} \mathbin{/} 1 & &%
    \SetCell[c=2]{c} t \mathbin{/} \unit{\second} & &%
    \SetCell[c=2]{c} F_\text{i} \mathbin{/} 1 & &%
    \SetCell[c=2]{c} F_\text{f} \mathbin{/} 1 & &%
    \SetCell[c=2]{c} t \mathbin{/} \unit{\second}& \\
    \midrule
"""


table_footer = r"""    \bottomrule
  \end{tblr}
"""


row_template = (
    r"{0.n:1.3f} & {1.n:0.0f} & {1.s:0.0f} &  {2.n:0.0f} & {2.s:0.0f}&  {3.n:1.3f} & {3.s:1.3f}  &  {4.n:0.0f} &  {4.s:0.0f}&  {5.n:0.0f} & {5.s:0.0f}&  {6.n:1.3f} & {6.s:1.3f}   \\"
)



with open("build/table_all-measurements.tex", "w") as f:
    f.write(table_header)
    for row in zip(h_b, Fi_b, Ff_b, t_b, Fi_c, Ff_c, t_c):
        f.write(row_template.format(*row))
        f.write("\n")
    f.write(table_footer)


# second: writing the averaged values and leaving out the frames
# heights_ball, durations_ball, durations_cylinder
# have not been reduced to 6 values each


table_header = r"""
  \begin{tblr}{
    colspec = {
        S[table-format=1.3]
        S[table-format=2.3]
        S[table-format=1.4]
        S[table-format=3.3]
        S[table-format=1.5]
    },
    row{1} = {guard},
    row{2} = {guard, mode=math},
    vline{3,5,7,9} = {2}{-}{text=\clap{\pm}},
  }
    \toprule
    &
    \SetCell[c=2]{c} Messung: Kugel & &
    \SetCell[c=2]{c} Messung: Zylinder &\\
    \cmidrule[lr]{2-3}\cmidrule[lr]{4-5}
    h \mathbin{/} \unit{\meter} &%
    \SetCell[c=2]{c} t \mathbin{/} \unit{\second} & &%
    \SetCell[c=2]{c} t \mathbin{/} \unit{\second}& \\
    \midrule
"""


table_footer = r"""    \bottomrule
  \end{tblr}
"""


row_template = (
    r"{0.n:1.3f} & {1.n:1.3f} & {1.s:1.3f} &  {2.n:1.3f} & {2.s:1.3f}   \\"
)


with open("build/table_averaged-measurements.tex", "w") as f:
    f.write(table_header)
    for row in zip(h_b_mean, t_b_mean, t_c_mean):
        f.write(row_template.format(*row))
        f.write("\n")
    f.write(table_footer)